import 'package:flutter/material.dart';
import 'package:mobile_ui/src/utils/function.dart';

import '../mobile_ui_container.dart';
import 'circular_loading.dart';

typedef ButtonBuilder = Widget Function(CustomButtonStyle style,
    VoidCallback? onPressed, bool loading, BuildContext context);

enum ButtonType {
  raised,
  flat,
  outline,
  text,
}

enum ButtonStyleType {
  primaryButton,
  secondaryButton,
  defaultButton,
  textButton
}

class CustomButtonStyle {
  Color? color;
  Color? textColor;
  ButtonType buttonType;
  double? height;
  double? fontSizeFactor;
  double? fontSizeDelta;
  EdgeInsetsGeometry? padding;
  double? radiusWidth;
  double? elevation;
  Color? shadowColor;
  Color? borderColor;

  CustomButtonStyle(
      {required this.buttonType,
      this.color,
      this.textColor,
      this.height,
      this.fontSizeFactor,
      this.fontSizeDelta,
      this.padding,
      this.radiusWidth,
      this.elevation,
      this.shadowColor,
      this.borderColor});

  CustomButtonStyle copyWith({double? elevation, Color? shadowColor}) {
    return CustomButtonStyle(
        buttonType: buttonType,
        color: color,
        textColor: textColor,
        height: height,
        fontSizeDelta: fontSizeDelta,
        fontSizeFactor: fontSizeFactor,
        padding: padding,
        radiusWidth: radiusWidth,
        elevation: elevation ?? this.elevation,
        shadowColor: shadowColor ?? this.shadowColor);
  }

  CustomButtonStyle apply(
      {Color? color,
      Color? textColor,
      double? height,
      double? fontSizeFactor,
      double? elevation,
      double? fontSizeDelta,
      EdgeInsetsGeometry? padding,
      double? radiusWidth,
      Color? shadowColor,
      Color? borderColor}) {
    this.color = color ?? this.color;
    this.textColor = textColor ?? this.textColor;
    this.height = height ?? this.height;
    this.fontSizeFactor = fontSizeFactor ?? this.fontSizeFactor;
    this.elevation = elevation ?? this.elevation;
    this.fontSizeDelta = fontSizeDelta ?? this.fontSizeDelta;
    this.padding = padding ?? this.padding;
    this.radiusWidth = radiusWidth ?? this.radiusWidth;
    this.borderColor = borderColor ?? this.borderColor;
    return this;
  }

  CustomButtonStyle.raised(
      {Color? color,
      Color? textColor,
      double? height,
      double? fontSizeFactor,
      double? elevation,
      double? fontSizeDelta,
      EdgeInsetsGeometry? padding,
      double? radiusWidth,
      Color? shadowColor})
      : this(
            buttonType: ButtonType.raised,
            color: color,
            textColor: textColor,
            height: height,
            fontSizeDelta: fontSizeDelta,
            fontSizeFactor: fontSizeFactor,
            padding: padding,
            radiusWidth: radiusWidth,
            elevation: elevation,
            shadowColor: shadowColor);

  CustomButtonStyle.flat(
      {Color? color,
      Color? textColor,
      double? height,
      double? fontSizeFactor,
      double? fontSizeDelta,
      EdgeInsetsGeometry? padding,
      double? radiusWidth})
      : this(
            buttonType: ButtonType.flat,
            color: color,
            textColor: textColor,
            height: height,
            fontSizeDelta: fontSizeDelta,
            fontSizeFactor: fontSizeFactor,
            padding: padding,
            radiusWidth: radiusWidth);

  CustomButtonStyle.outline(
      {Color? color,
      Color? textColor,
      double? height,
      double? fontSizeFactor,
      double? fontSizeDelta,
      EdgeInsetsGeometry? padding,
      double? radiusWidth,
      Color? borderColor})
      : this(
            buttonType: ButtonType.outline,
            color: color,
            textColor: textColor,
            height: height,
            fontSizeDelta: fontSizeDelta,
            fontSizeFactor: fontSizeFactor,
            padding: padding,
            radiusWidth: radiusWidth,
            borderColor: borderColor);

  CustomButtonStyle.text(
      {Color? color,
      Color? textColor,
      double? height,
      double? fontSizeFactor,
      double? fontSizeDelta,
      EdgeInsetsGeometry? padding})
      : this(
            buttonType: ButtonType.outline,
            color: color,
            textColor: textColor,
            height: height,
            fontSizeDelta: fontSizeDelta,
            fontSizeFactor: fontSizeFactor,
            padding: padding);
}

class Button extends StatefulWidget {
  final CustomButtonStyle? buttonStyle;
  final VoidCallback? onPressed;
  final AsyncVoidCallback? onAsyncPressed;
  final String? text;
  final String? loadingText;
  final double? fontSizeFactor;
  final double? fontSizeDelta;
  final IconData? icon;
  final TextAlign textAlign;
  final MainAxisAlignment align;
  final bool? isShowLoading;
  final double? height;
  final double? width;
  final ButtonStyleType? type;

  const Button({
    this.onPressed,
    this.onAsyncPressed,
    this.buttonStyle,
    this.text,
    this.loadingText,
    this.fontSizeFactor,
    this.fontSizeDelta,
    this.icon,
    this.textAlign = TextAlign.center,
    this.align = MainAxisAlignment.center,
    this.isShowLoading,
    this.height,
    this.width,
    this.type = ButtonStyleType.primaryButton,
    Key? key,
  }) : super(key: key);

  @override
  _ButtonState createState() => _ButtonState();
}

class _ButtonState extends State<Button> {
  var _loading = false;

  @override
  Widget build(BuildContext context) {
    ThemeData theme = Theme.of(context);
    CustomButtonStyle style = widget.buttonStyle ??
        _getConfigStyle(context) ??
        CustomButtonStyle.flat(color: theme.primaryColor, radiusWidth: 2.0);
    switch (style.buttonType) {
      case ButtonType.text:
        return _buildButton(context, style, _buildTextButton);
      case ButtonType.outline:
        return _buildButton(context, style, _buildOutlineButton);
      case ButtonType.flat:
        return _buildButton(context, style, _buildElevatedButton);
      case ButtonType.raised:
      default:
        return _buildButton(
            context,
            style.copyWith(
                elevation: style.elevation ?? 10,
                shadowColor: style.shadowColor ?? theme.shadowColor),
            _buildElevatedButton);
    }
  }

  CustomButtonStyle? _getConfigStyle(BuildContext context) {
    final uiConfig = MobileUIContainer.of(context);
    if (uiConfig == null) return null;
    switch (widget.type) {
      case ButtonStyleType.defaultButton:
        return uiConfig.defaultButtonStyle;
      case ButtonStyleType.secondaryButton:
        return uiConfig.secondaryButtonStyle;
      case ButtonStyleType.textButton:
        return uiConfig.textButtonStyle;
      default:
        return uiConfig.primaryButtonStyle;
    }
  }

  Widget _buildButton(BuildContext context, CustomButtonStyle style,
      ButtonBuilder buttonBuilder) {
    final showLoading = widget.isShowLoading ?? (widget.onAsyncPressed != null);
    if (showLoading) {
      return buttonBuilder(
          style,
          (widget.onPressed == null && widget.onAsyncPressed == null)
              ? null
              : () async {
                  if (_loading) return;
                  setState(() {
                    _loading = true;
                  });
                  await _handleFocus();
                  if (widget.onPressed != null) {
                    widget.onPressed!();
                  }
                  if (widget.onAsyncPressed != null) {
                    await widget.onAsyncPressed!();
                  }
                  setState(() {
                    _loading = false;
                  });
                },
          _loading,
          context);
    } else {
      return buttonBuilder(
          style,
          (widget.onPressed == null && widget.onAsyncPressed == null)
              ? null
              : () async {
                  await _handleFocus();
                  if (widget.onPressed != null) {
                    widget.onPressed!();
                  }
                  if (widget.onAsyncPressed != null) {
                    await widget.onAsyncPressed!();
                  }
                },
          false,
          context);
    }
  }

  Future<void> _handleFocus() async {
    final effectiveFocusNode = FocusManager.instance.primaryFocus;
    if (effectiveFocusNode != null && effectiveFocusNode.hasFocus) {
      effectiveFocusNode.unfocus();
      await Future.delayed(Duration(milliseconds: 20));
    }
  }

  Widget _buildTextButton(CustomButtonStyle style, VoidCallback? onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    final textWidget = _buildTextWidget(loading, theme, onPressed, style);
    return Row(
      mainAxisAlignment: widget.align,
      children: [
        if (widget.icon != null)
          Icon(widget.icon,
              color: onPressed == null ? theme.disabledColor : style.textColor),
        TextButton(onPressed: onPressed, child: textWidget),
      ],
    );
  }

  Text _buildTextWidget(bool loading, ThemeData theme, VoidCallback? onPressed,
      CustomButtonStyle style) {
    return Text(
      (loading && widget.loadingText != null && widget.loadingText!.isNotEmpty)
          ? widget.loadingText!
          : widget.text!,
      textAlign: widget.textAlign,
      style: theme.textTheme.button!.apply(
          color: onPressed != null ? style.textColor : Colors.black26,
          fontSizeFactor: widget.fontSizeFactor ?? style.fontSizeFactor ?? 1.0,
          fontSizeDelta: widget.fontSizeDelta ?? style.fontSizeDelta ?? 0.0),
    );
  }

  Widget _buildOutlineButton(CustomButtonStyle style, VoidCallback? onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    style.apply(
        textColor: style.textColor ?? theme.primaryColor,
        borderColor: style.borderColor ?? theme.primaryColor);
    final textWidget = _buildTextWidget(loading, theme, onPressed, style);
    final ButtonStyle outlineButtonStyle = OutlinedButton.styleFrom(
      primary: style.color,
      fixedSize: Size(
          widget.width ?? double.infinity, widget.height ?? style.height ?? 36),
      side: BorderSide(color: style.borderColor ?? Colors.black26),
      padding: style.padding,
      shape: RoundedRectangleBorder(
        borderRadius:
            BorderRadius.all(Radius.circular(style.radiusWidth ?? 2.0)),
      ),
    );
    return OutlinedButton(
      onPressed: onPressed,
      style: outlineButtonStyle,
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          if (loading)
            CircularLoading(
              color: style.textColor,
            ),
          if (widget.icon != null)
            Icon(widget.icon,
                color:
                    onPressed == null ? theme.disabledColor : style.textColor),
          textWidget
        ],
      ),
    );
  }

  Widget _buildElevatedButton(CustomButtonStyle style, VoidCallback? onPressed,
      bool loading, BuildContext context) {
    final theme = Theme.of(context);
    final textWidget = _buildTextWidget(loading, theme, onPressed, style);
    final ButtonStyle raisedButtonStyle = ElevatedButton.styleFrom(
      onPrimary: style.textColor,
      elevation: style.elevation,
      shadowColor: style.shadowColor,
      primary: style.color,
      fixedSize: Size(
          widget.width ?? double.infinity, widget.height ?? style.height ?? 36),
      padding: style.padding,
      shape: RoundedRectangleBorder(
        borderRadius:
            BorderRadius.all(Radius.circular(style.radiusWidth ?? 2.0)),
      ),
    );

    return ElevatedButton(
      onPressed: onPressed,
      style: raisedButtonStyle,
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          if (loading)
            CircularLoading(
              color: style.textColor,
            ),
          if (widget.icon != null)
            Icon(widget.icon,
                color:
                    onPressed == null ? theme.disabledColor : style.textColor),
          textWidget
        ],
      ),
    );
  }
}
