import 'package:flutter/material.dart';
import 'package:mobile_ui/mobile_ui.dart';
import 'package:mobile_ui/src/controls/spacing.dart';

import '../mobile_ui_container.dart';

typedef ValueValidator<T> = String? Function(
    T value, String? errorCode, String? errorMsg);
enum ControlPosition { leading, trailing }
enum FieldOrientation { vertical, horizontal }

class FormInputLabelStyle {
  const FormInputLabelStyle(
      {this.color, this.maxLine, this.fontSize, this.fontFamily});

  final Color? color;
  final int? maxLine;
  final double? fontSize;
  final String? fontFamily;
}

class FormInputTextStyle {
  const FormInputTextStyle(
      {this.color, this.disabledColor, this.fontSize, this.fontFamily});

  final Color? color;
  final Color? disabledColor;
  final double? fontSize;
  final String? fontFamily;
}

class FormInputBackgroundStyle {
  const FormInputBackgroundStyle(
      {this.backgroundColor,
      this.borderColor,
      this.borderWidth,
      this.disabledBackgroundColor,
      this.disabledBorderColor});

  static const FormInputBackgroundStyle none = FormInputBackgroundStyle();

  FormInputBackgroundStyle merge(FormInputBackgroundStyle style) {
    return FormInputBackgroundStyle(
        backgroundColor: backgroundColor ?? style.backgroundColor,
        borderColor: borderColor ?? style.borderColor,
        borderWidth: borderWidth ?? style.borderWidth,
        disabledBackgroundColor:
            disabledBackgroundColor ?? style.disabledBackgroundColor,
        disabledBorderColor: disabledBorderColor ?? style.disabledBorderColor);
  }

  final Color? backgroundColor;
  final Color? borderColor;
  final double? borderWidth;
  final Color? disabledBackgroundColor;
  final Color? disabledBorderColor;
}

const disabledBackgroundColor = Color(0xFFe5e5e5);
const disabledBorderColor = Color(0xFFd5d5d5);
const disabledTextColor = Color(0xFF838383);

const FormInputBackgroundStyle defaultFormInputBackgroundStyle =
    FormInputBackgroundStyle(backgroundColor: Colors.white);

const FormInputLabelStyle defaultFormInputLabelStyle =
    FormInputLabelStyle(color: Colors.black);

const FormInputTextStyle defaultFormInputTextStyle =
    FormInputTextStyle(color: Colors.black);

class FormInputBase<T> extends FormField<T> {
  final bool? isRequired;
  final String? labelText;
  final String? hintText;
  final EdgeInsetsGeometry contentPadding;
  final EdgeInsetsGeometry margin;
  final double spacingFlactor;
  final bool enabled;
  final bool readOnly;
  final TextAlign labelAlign;
  final FieldOrientation orientation;
  final ControlPosition controlPosition;
  final ValueNotifier<T?>? valueController;
  final ValueChanged<T?>? onChanged;
  final Icon? prefixIcon;
  final bool? isDense;
  final int errorMaxLines;
  final double? labelWidth;
  final FormInputLabelStyle? labelStyle;
  final FormInputBackgroundStyle? backgroundStyle;
  final FormInputTextStyle? textStyle;

  FormInputBase(
      {Key? key,
      required FormFieldBuilder<T> builder,
      ValueValidator<T?>? validator,
      T? initialValue,
      this.valueController,
      this.prefixIcon,
      this.isRequired,
      this.labelText,
      this.hintText,
      this.contentPadding = const EdgeInsets.all(8),
      this.margin = const EdgeInsets.symmetric(vertical: 8),
      this.spacingFlactor = 1,
      required bool enabled,
      this.isDense,
      bool? readOnly,
      this.labelAlign = TextAlign.left,
      this.orientation = FieldOrientation.vertical,
      //Only valid when the direction is equal to horizontal
      this.controlPosition = ControlPosition.trailing,
      AutovalidateMode? autovalidateMode,
      FormFieldSetter<T>? onSaved,
      this.onChanged,
      this.errorMaxLines = 1,
      this.labelWidth,
      this.labelStyle,
      this.backgroundStyle,
      this.textStyle})
      : assert(initialValue == null || valueController == null),
        enabled = enabled,
        readOnly = (enabled == false || readOnly == true),
        super(
            builder: builder,
            initialValue:
                valueController != null ? valueController.value : initialValue,
            onSaved: onSaved,
            enabled: readOnly == true,
            autovalidateMode: autovalidateMode,
            validator: (value) {
              String? errorCode;
              String? errorMsg;
              final isNullOrEmpty = (value == null ||
                  (value is String && value.isEmpty) ||
                  (value is List && value.isEmpty));
              if (isNullOrEmpty) {
                if (isRequired == true) {
                  errorCode = 'required';
                  errorMsg = '${labelText ?? "This field"} is required.';
                }
              }
              if (validator != null)
                return validator(value, errorCode, errorMsg);
              return errorMsg;
            });

  Widget buildField(FormFieldState<T> field, Widget child) {
    final theme = Theme.of(field.context);
    final uiConfig = MobileUIContainer.of(field.context);
    final bgStyle = backgroundStyle ??
        (uiConfig?.formInputBackgroundStyle ?? defaultFormInputBackgroundStyle);
    final formLabelStyle = labelStyle ??
        uiConfig?.formInputLabelStyle ??
        defaultFormInputLabelStyle;
    if (orientation == FieldOrientation.vertical) {
      return buildVField(
          theme, buildDecorator(field, child, bgStyle), formLabelStyle);
    } else {
      return buildHField(
          theme, buildDecorator(field, child, bgStyle), formLabelStyle);
    }
  }

  Widget buildHField(
      ThemeData theme, Widget child, FormInputLabelStyle formLabelStyle) {
    final hasLabel = (labelText != null && labelText!.isNotEmpty);
    var children = <Widget>[];
    if (hasLabel) {
      if (controlPosition == ControlPosition.leading) {
        children.addAll([
          Expanded(child: child),
          Spacing(spacingFlactor),
          buildLabelText(theme, formLabelStyle)
        ]);
      } else {
        children.addAll([
          buildLabelText(theme, formLabelStyle),
          Spacing(spacingFlactor),
          Expanded(child: child)
        ]);
      }
    }
    return Padding(
        padding: margin,
        child: hasLabel
            ? Row(
                children: children,
              )
            : child);
  }

  Widget buildVField(
      ThemeData theme, Widget child, FormInputLabelStyle formLabelStyle) {
    final hasLabel = (labelText != null && labelText!.isNotEmpty);
    var children = <Widget>[];
    if (hasLabel) {
      children.addAll([
        buildLabelText(theme, formLabelStyle),
        Spacing(spacingFlactor),
      ]);
    }
    return Padding(
        padding: margin,
        child: hasLabel
            ? Column(
                crossAxisAlignment: CrossAxisAlignment.stretch,
                children: [...children, child])
            : child);
  }

  Widget buildLabelText(ThemeData theme, FormInputLabelStyle formLabelStyle) {
    late Widget labelWidget;
    final labelColor = formLabelStyle.color;
    final labelFontSize = formLabelStyle.fontSize;
    final labelFontFamily = formLabelStyle.fontFamily;
    final text = Text(
      labelText ?? '',
      textAlign: labelAlign,
      maxLines: formLabelStyle.maxLine,
      style: enabled
          ? theme.textTheme.subtitle1!.copyWith(
              color: labelColor,
              fontSize: labelFontSize,
              fontFamily: labelFontFamily)
          : theme.textTheme.subtitle1!.copyWith(
              color: theme.disabledColor,
              fontSize: labelFontSize,
              fontFamily: labelFontFamily),
    );
    if (isRequired == true) {
      var mainAxisAlignment = MainAxisAlignment.start;
      if (labelAlign == TextAlign.center) {
        mainAxisAlignment = MainAxisAlignment.center;
      } else if (labelAlign == TextAlign.right) {
        mainAxisAlignment = MainAxisAlignment.end;
      }
      labelWidget = Row(
        mainAxisAlignment: mainAxisAlignment,
        mainAxisSize: MainAxisSize.min,
        children: [
          Text(
            '* ',
            textAlign: labelAlign,
            style: enabled
                ? theme.textTheme.subtitle1!.copyWith(color: Colors.red)
                : theme.textTheme.subtitle1!
                    .copyWith(color: theme.disabledColor),
          ),
          text,
        ],
      );
    } else {
      labelWidget = text;
    }
    if (labelWidth != null) {
      return SizedBox(width: labelWidth, child: labelWidget);
    }
    return labelWidget;
  }

  TextStyle getTextStyle(FormFieldState<T> field) {
    final theme = Theme.of(field.context);
    final uiConfig = MobileUIContainer.of(field.context);
    final txtStyle = textStyle ??
        (uiConfig?.formInputTextStyle ?? defaultFormInputTextStyle);
    return theme.textTheme.subtitle1!.copyWith(
        color: readOnly
            ? txtStyle.disabledColor ?? disabledTextColor
            : txtStyle.color,
        fontFamily: txtStyle.fontFamily,
        fontSize: txtStyle.fontSize);
  }

  Widget buildText(FormFieldState<T> field, String? text, {int? maxLines}) {
    return SelectableText(
      text ?? '',
      enableInteractiveSelection: enabled != false,
      textAlign: TextAlign.left,
      maxLines: maxLines,
      style: getTextStyle(field),
    );
  }

  Widget buildIcon(ThemeData theme, IconData icon,
      {GestureTapCallback? onTap, Color? enabledColor}) {
    final color = (enabled && !readOnly)
        ? enabledColor ?? theme.primaryColor
        : theme.disabledColor;
    final iconWidget = Icon(
      icon,
      color: color,
    );
    if (onTap != null && !readOnly) {
      return InkWell(
        onTap: onTap,
        child: iconWidget,
      );
    }
    return iconWidget;
  }

  bool getIsNullOrEmpty(T? value) {
    return (value == null ||
        (value is String && value.isEmpty) ||
        (value is List && value.isEmpty));
  }

  Widget buildDecorator(FormFieldState<T> field, Widget child,
      FormInputBackgroundStyle backgroundStyle) {
    final effectiveDecoration =
        InputDecoration(contentPadding: EdgeInsets.all(0)).copyWith(
            hintText: hintText,
            border: OutlineInputBorder(borderSide: BorderSide.none));
    final isNullOrEmpty = getIsNullOrEmpty(field.value);

    if (backgroundStyle == FormInputBackgroundStyle.none) {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              hintText: hintText,
              errorText: field.errorText,
              errorMaxLines: errorMaxLines,
              isDense: isDense),
          isEmpty: isNullOrEmpty,
          child: child);
    }

    final hasBorder = backgroundStyle.borderWidth != null;
    final theme = Theme.of(field.context);
    final backgroundColor = readOnly
        ? backgroundStyle.disabledBackgroundColor ?? disabledBackgroundColor
        : backgroundStyle.backgroundColor;

    final borderColor = readOnly
        ? backgroundStyle.disabledBorderColor ?? disabledBorderColor
        : backgroundStyle.borderColor;

    if (hasBorder) {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              fillColor: backgroundColor,
              filled: backgroundColor != null,
              prefixIcon: prefixIcon,
              errorText: field.errorText,
              errorMaxLines: errorMaxLines,
              isDense: isDense,
              contentPadding: contentPadding,
              hintText: hintText,
              errorBorder: OutlineInputBorder(
                  borderSide: BorderSide(
                      color: theme.errorColor,
                      width: backgroundStyle.borderWidth ?? 1)),
              enabledBorder: OutlineInputBorder(
                  borderSide: BorderSide(
                      color: borderColor ?? Color(0xFFEEEEEE),
                      width: backgroundStyle.borderWidth ?? 1))),
          isEmpty: isNullOrEmpty,
          child: child);
    } else {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              hintText: hintText,
              errorText: field.errorText,
              errorMaxLines: errorMaxLines,
              isDense: isDense),
          isEmpty: isNullOrEmpty,
          child: Container(
              constraints: BoxConstraints(minHeight: 48),
              color: readOnly
                  ? backgroundStyle.disabledBackgroundColor ??
                      disabledBackgroundColor
                  : backgroundStyle.backgroundColor,
              padding: contentPadding,
              child: child));
    }
  }

  @override
  FormFieldState<T> createState() => FormInputStateBase<T>();
}

class FormInputStateBase<T> extends FormFieldState<T> {
  ValueNotifier<T?>? _valueController;

  ValueNotifier<T?>? get effectiveValueController =>
      widget.valueController ?? _valueController;

  @override
  FormInputBase<T> get widget => super.widget as FormInputBase<T>;

  @override
  void initState() {
    super.initState();
    if (widget.valueController == null) {
      _valueController = ValueNotifier<T?>(widget.initialValue);
    } else {
      widget.valueController!.addListener(handleControllerChanged);
    }
  }

  @override
  bool validate() {
    saveValue();
    return super.validate();
  }

  @override
  void didUpdateWidget(FormInputBase<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.valueController != oldWidget.valueController) {
      oldWidget.valueController?.removeListener(handleControllerChanged);
      widget.valueController?.addListener(handleControllerChanged);

      if (oldWidget.valueController != null && widget.valueController == null) {
        _valueController = ValueNotifier<T?>(oldWidget.valueController!.value);
      }
      if (widget.valueController != null) {
        setValue(widget.valueController!.value);
        if (oldWidget.valueController == null) _valueController = null;
      }
    }
  }

  @override
  void save() {
    saveValue();
    super.save();
  }

  @override
  void dispose() {
    widget.valueController?.removeListener(handleControllerChanged);
    super.dispose();
  }

  @override
  void reset() {
    super.reset();
    //setState(() {
    effectiveValueController!.value = widget.initialValue;
    //});
  }

  @protected
  void handleControllerChanged() {
    changeValue(effectiveValueController!.value);
  }

  @protected
  void saveValue() {
    // changeValue(effectiveValueController!.value);
  }

  void changeValue(T? v) {
    if (effectiveValueController!.value != v) {
      effectiveValueController!.value = v;
    }
    if (v != value) {
      didChange(v);
      if (widget.onChanged != null) {
        widget.onChanged!(v);
      }
    }
  }
}

class ValueController<T> extends ValueNotifier<T> {
  ValueController(T value) : super(value);

  ValueController.fromValue(T value) : super(value);
}
