import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

import 'form_input_base.dart';

class NumFormInput extends FormInputBase<num> {
  NumFormInput(
      {Key? key,
      this.step,
      this.fractionDigits = 0,
      this.minValue = 0,
      this.maxValue,
      ValueController<num?>? valueController,
      String? labelText,
      String? hintText,
      num? initialValue,
      this.focusNode,
      bool readOnly = false,
      bool enabled = true,
      bool isRequired = false,
      bool? isDense,
      FormInputBackgroundStyle? backgroundStyle,
      TextInputAction? textInputAction,
      TextAlign textAlign = TextAlign.left,
      VoidCallback? onEditingComplete,
      FormFieldSetter<num>? onSaved,
      ValueValidator<num?>? validator,
      ValueChanged<num?>? onChanged,
      AutovalidateMode? autovalidateMode,
      TextAlign labelAlign = TextAlign.left,
      double spacingFlactor = 1,
      FieldOrientation orientation = FieldOrientation.vertical,
      EdgeInsetsGeometry contentPadding = const EdgeInsets.all(8),
      EdgeInsetsGeometry margin = const EdgeInsets.symmetric(vertical: 8),
      ControlPosition controlPosition = ControlPosition.trailing,
      double? labelWidth,
      Widget? suffix,
      int errorMaxLines = 1,
      this.includeMinValue = true,
      this.includeMaxValue = true,
      FormInputLabelStyle? labelStyle,
      FormInputTextStyle? textStyle})
      : assert(initialValue == null || valueController == null),
        super(
          initialValue: initialValue,
          labelWidth: labelWidth,
          onSaved: onSaved,
          isDense: isDense,
          onChanged: onChanged,
          labelStyle: labelStyle,
          backgroundStyle: backgroundStyle,
          textStyle: textStyle,
          validator: (value, errorCode, errorMsg) {
            var eCode = errorCode;
            var eMsg = errorMsg;
            if (errorCode == null || errorCode.isEmpty) {
              if (value != null) {
                if (includeMinValue == true && minValue > value) {
                  eCode = 'cantLessThanMinValue';
                  eMsg =
                      '${labelText ?? "This field"} cannot be less than $minValue.';
                } else if (includeMinValue != true && minValue >= value) {
                  eCode = 'mustGreaterThanMinValue';
                  eMsg =
                      '${labelText ?? "This field"} must be greater than $minValue.';
                } else if (includeMaxValue == true &&
                    maxValue != null &&
                    maxValue < value) {
                  eCode = 'cantGreaterThanMaxValue';
                  eMsg =
                      '${labelText ?? "This field"} cannot be greater than $maxValue.';
                } else if (includeMaxValue != true &&
                    maxValue != null &&
                    maxValue <= value) {
                  eCode = 'mustLessThanMaxValue';
                  eMsg =
                      '${labelText ?? "This field"} must be less than $maxValue.';
                }
              }
            }
            if (validator != null) return validator(value, eCode, eMsg);
            return eMsg;
          },
          readOnly: readOnly,
          spacingFlactor: spacingFlactor,
          enabled: enabled,
          autovalidateMode: autovalidateMode,
          orientation: orientation,
          controlPosition: controlPosition,
          margin: margin,
          labelText: labelText,
          labelAlign: labelAlign,
          valueController: valueController,
          isRequired: isRequired,
          contentPadding: contentPadding,
          errorMaxLines: errorMaxLines,
          builder: (field) {
            var widget = field.widget as NumFormInput;
            final state = field as _CounterFormFieldState;
            final theme = Theme.of(field.context);
            final whitelistChar = minValue >= 0 ? "[0-9.]" : "[0-9.\-]";
            final child = Row(
              children: <Widget>[
                Expanded(
                  child: TextField(
                    inputFormatters: [
                      FilteringTextInputFormatter.allow(RegExp(whitelistChar)),
                    ],
                    controller: state._textEditingController,
                    focusNode: state.effectiveFocusNode,
                    decoration: null,
                    keyboardType: TextInputType.numberWithOptions(
                        signed: minValue < 0, decimal: fractionDigits > 0),
                    textInputAction: textInputAction,
                    textAlign: textAlign,
                    textCapitalization: TextCapitalization.none,
                    autofocus: false,
                    autocorrect: true,
                    onEditingComplete: onEditingComplete,
                    enabled: enabled,
                    readOnly: readOnly,
                    style: widget.getTextStyle(field),
                  ),
                ),
                if (suffix != null) suffix,
                if (step != null)
                  widget.buildIcon(theme, Icons.remove, onTap: () async {
                    await state.unFocus();
                    final val = (state.value ?? 0) - step;
                    state.changeValue(val);
                    state.changeText();
                  }),
                if (step != null)
                  widget.buildIcon(theme, Icons.add, onTap: () async {
                    await state.unFocus();
                    final val = (state.value ?? 0) + step;
                    state.changeValue(val);
                    state.changeText();
                  }),
              ],
            );
            return widget.buildField(field, child);
          },
        );

  /// Controls the text being edited.
  ///
  /// If null, this widget will create its own [TextEditingController] and
  /// initialize its [TextEditingController.text] with [initialValue].
  final num? step;
  final int fractionDigits;
  final num minValue;
  final num? maxValue;
  final bool? includeMinValue;
  final bool? includeMaxValue;
  final FocusNode? focusNode;

  @override
  _CounterFormFieldState createState() => _CounterFormFieldState();
}

class _CounterFormFieldState extends FormInputStateBase<num> {
  TextEditingController? _textEditingController;
  FocusNode? _focusNode;

  @protected
  FocusNode? get effectiveFocusNode => widget.focusNode ?? _focusNode;

  @override
  void initState() {
    super.initState();
    _textEditingController = TextEditingController(text: getText());
    if (widget.focusNode == null) {
      _focusNode = FocusNode();
    }

    effectiveFocusNode!.addListener(_focusChanged);
  }

  void _focusChanged() {
    if (effectiveFocusNode!.hasFocus) {
      if (_textEditingController!.text.isNotEmpty)
        _textEditingController!.selection = TextSelection(
            baseOffset: 0, extentOffset: _textEditingController!.text.length);
    } else {
      _setTextValue();
    }
  }

  @override
  void didUpdateWidget(NumFormInput oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.focusNode != oldWidget.focusNode) {
      oldWidget.focusNode?.removeListener(_focusChanged);
      widget.focusNode?.addListener(_focusChanged);

      if (oldWidget.focusNode != null && widget.focusNode == null) {
        _focusNode = FocusNode();
      }
      if (widget.focusNode != null) {
        if (oldWidget.focusNode == null) _focusNode = null;
      }
    }
  }

  @override
  void reset() {
    super.reset();
    changeValue(effectiveValueController!.value);
  }

  void _setTextValue() {
    final text = _textEditingController!.text;
    num? v;
    if (text.isNotEmpty) {
      v = num.tryParse(text);
      v = v ?? value;
    }
    changeValue(getValidValue(v));
  }

  String getText() {
    return value?.toStringAsFixed(widget.fractionDigits) ?? '';
  }

  void changeText() {
    final text = getText();
    if (_textEditingController!.text != text) {
      _textEditingController!.text = text;
    }
  }

  @override
  NumFormInput get widget => super.widget as NumFormInput;

  @override
  void dispose() {
    widget.focusNode?.removeListener(_focusChanged);
    _focusNode?.removeListener(_focusChanged);
    _focusNode?.dispose();
    _textEditingController!.dispose();
    super.dispose();
  }

  num? getValidValue(num? v) {
    if (v != null && v < widget.minValue) v = widget.minValue;
    if (v != null && widget.maxValue != null && v > widget.maxValue!)
      v = widget.maxValue!;
    return v;
  }

  @override
  void saveValue() {
    _setTextValue();
  }

  @override
  void changeValue(num? v) {
    super.changeValue(getValidValue(v));
    final text = getText();
    if (_textEditingController!.text != text) {
      _textEditingController!.text = text;
    }
  }

  Future<void> unFocus() async {
    if (effectiveFocusNode != null && effectiveFocusNode!.hasFocus) {
      effectiveFocusNode!.unfocus();
      await Future.delayed(Duration(milliseconds: 20));
    }
  }
}
//
//class NumericTextFormatter extends TextInputFormatter {
//  NumericTextFormatter({this.fractionDigits = 0});
//
//  int fractionDigits;
//
//  TextEditingValue formatEditUpdate(
//      TextEditingValue oldValue, TextEditingValue newValue) {
//    if (newValue.text.isEmpty) {
//      return newValue.copyWith(text: '');
//    } else if (newValue.text.compareTo(oldValue.text) != 0) {
//      var selectionIndexFromTheRight =
//          newValue.text.length - newValue.selection.end;
//      var s = '';
//      for (var i = 0; i < fractionDigits; i++) {
//        s += '#';
//      }
//      final f = NumberFormat("#,###${s.isNotEmpty ? '.$s' : ''}");
//      var num = int.parse(newValue.text.replaceAll(f.symbols.GROUP_SEP, ''));
//      final newString = f.format(num);
//      return TextEditingValue(
//        text: newString,
//        selection: TextSelection.collapsed(
//            offset: newString.length - selectionIndexFromTheRight),
//      );
//    } else {
//      return newValue;
//    }
//  }
//}
