import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:validators/validators.dart';

import '../../mobile_ui.dart';
import 'form_input_base.dart';

enum TextType { text, password, email, url, phone, ip, money }

class TextFormInput extends FormInputBase<String> {
  final FocusNode? focusNode;
  final Icon? prefixIcon;
  final TextEditingController? textController;
  final TextType textType;

  TextFormInput(
      {Key? key,
      String? initialValue,
      this.textController,
      String? labelText,
      double? labelWidth,
      String? hintText,
      int? minLength,
      int? maxLength,
      bool isRequired = false,
      bool? isShowCounter,
      bool? isDense,
      this.textType = TextType.text,
      bool readOnly = false,
      bool enabled = true,
      FormFieldSetter<String>? onSaved,
      ValueChanged<String?>? onChanged,
      ValueChanged<String>? onSubmitted,
      VoidCallback? onEditingComplete,
      this.focusNode,
      TextInputAction? textInputAction,
      this.prefixIcon,
      bool autofocus = false,
      ValueValidator<String?>? validator,
      int? minLines,
      int? maxLines = 1,
      TextAlign textAlign = TextAlign.start,
      AutovalidateMode? autovalidateMode,
      TextAlign labelAlign = TextAlign.left,
      FieldOrientation orientation = FieldOrientation.vertical,
      EdgeInsetsGeometry contentPadding = const EdgeInsets.all(8),
      EdgeInsetsGeometry margin = const EdgeInsets.symmetric(vertical: 8),
      ControlPosition controlPosition = ControlPosition.trailing,
      double spacingFlactor = 1,
      int errorMaxLines = 1,
      List<TextInputFormatter>? inputFormatters,
      FormInputBackgroundStyle? backgroundStyle,
      FormInputLabelStyle? labelStyle,
      FormInputTextStyle? textStyle})
      : assert(initialValue == null || textController == null),
        assert(maxLines == null || maxLines >= 1),
        super(
          key: key,
          labelWidth: labelWidth,
          isDense: isDense,
          hintText: hintText,
          prefixIcon: prefixIcon,
          errorMaxLines: errorMaxLines,
          spacingFlactor: spacingFlactor,
          initialValue:
              textController == null ? initialValue : textController.text,
          onSaved: onSaved,
          validator: (value, errorCode, errorMsg) {
            String? eCode = errorCode;
            String? eMsg = errorMsg;
            if (value != null && (errorCode == null || errorCode.isEmpty)) {
              if (minLength != null && value.length < minLength) {
                eCode = 'lengthNotEnough';
                eMsg =
                    "${labelText ?? "This field"} cannot have less than $minLength characters.";
              } else if (maxLength != null && value.length > maxLength) {
                eCode = 'lengthExceeded';
                eMsg =
                    '${labelText ?? "This field"} cannot have more than $maxLength characters.';
              } else if (textType == TextType.email && !isEmail(value)) {
                eCode = 'invalidEmail';
                eMsg =
                    '${labelText ?? "This field"} requires a valid email address.';
              } else if (textType == TextType.url && !isURL(value)) {
                eCode = 'invalidUrl';
                eMsg =
                    '${labelText ?? "This field"} requires a valid url address.';
              } else if (textType == TextType.ip && !isIP(value)) {
                eCode = 'invalidIp';
                eMsg =
                    '${labelText ?? "This field"} requires a valid ip address.';
              } else if (textType == TextType.phone && !isNumeric(value)) {
                eCode = 'invalidPhone';
                eMsg =
                    '${labelText ?? "This field"} requires a valid phone number.';
              }
            }
            if (validator != null) return validator(value, eCode, eMsg);
            return eMsg;
          },
          enabled: enabled,
          readOnly: readOnly,
          autovalidateMode: autovalidateMode,
          orientation: orientation,
          controlPosition: controlPosition,
          labelText: labelText,
          labelAlign: labelAlign,
          isRequired: isRequired,
          contentPadding: contentPadding,
          margin: margin,
          onChanged: onChanged,
          backgroundStyle: backgroundStyle,
          labelStyle: labelStyle,
          textStyle: textStyle,
          builder: (field) {
            TextFormInput widget = field.widget as TextFormInput;
            final _TextFormFieldState state = field as _TextFormFieldState;
            final theme = Theme.of(field.context);
            TextInputType keyboardType;
            switch (textType) {
              case TextType.password:
                {
                  keyboardType = TextInputType.visiblePassword;
                }
                break;
              case TextType.email:
                {
                  keyboardType = TextInputType.emailAddress;
                }
                break;
              case TextType.url:
                {
                  keyboardType = TextInputType.url;
                }
                break;
              case TextType.phone:
                {
                  keyboardType = TextInputType.phone;
                }
                break;
              case TextType.ip:
                {
                  keyboardType = TextInputType.number;
                }
                break;
              case TextType.money:
                {
                  keyboardType = TextInputType.numberWithOptions(
                      decimal: true, signed: true);
                }
                break;
              default:
                {
                  keyboardType = TextInputType.text;
                }
                break;
            }
            Widget? builtCounter;
            bool _isShowCounter =
                isShowCounter ?? (maxLines != null && maxLines > 1);
            if (maxLength != null && _isShowCounter) {
              var style = enabled
                  ? theme.textTheme.caption
                  : theme.textTheme.caption!
                      .copyWith(color: theme.disabledColor);
              builtCounter = Align(
                alignment: Alignment.bottomRight,
                child: Text(
                  "${state._effectiveController!.text.length} / $maxLength",
                  style: style,
                ),
              );
            }
            Widget child;
            if (maxLines!=null && maxLines > 1) {
              keyboardType = TextInputType.multiline;
            }
            final textField = Row(children: <Widget>[
              Expanded(
                  child: TextField(
                controller: state._effectiveController,
                style: widget.getTextStyle(field),
                autofocus: autofocus,
                focusNode: focusNode,
                maxLines: maxLines,
                minLines: minLines,
                obscureText: state.obscureText,
                readOnly: readOnly,
                enabled: enabled,
                decoration: null,
                inputFormatters: inputFormatters,
                textInputAction: textInputAction ??
                    (maxLines!=null && maxLines > 1
                        ? TextInputAction.newline
                        : TextInputAction.next),
                keyboardType: keyboardType,
                textAlign: textAlign,
                textCapitalization: TextCapitalization.none,
                autocorrect: true,
                onEditingComplete: onEditingComplete,
                onSubmitted: onSubmitted,
                onChanged: (String value) {
                  field.didChange(value);
                },
                maxLength: maxLength,
              )),
              if (textType == TextType.password &&
                  state._effectiveController!.text.isNotEmpty)
                widget.buildIcon(
                    theme,
                    state.obscureText == true
                        ? Icons.visibility_off
                        : Icons.visibility, onTap: () {
                  if (!enabled) return;
                  field.setObscureText(!state.obscureText);
                }),
            ]);
            if (builtCounter != null)
              child = Column(
                children: [textField, builtCounter],
              );
            else
              child = textField;

            return widget.buildField(field, child);
          },
        );

  @override
  _TextFormFieldState createState() => _TextFormFieldState();
}

class _TextFormFieldState extends FormFieldState<String> {
  TextEditingController? _controller;
  bool obscureText = false;

  TextEditingController? get _effectiveController =>
      widget.textController ?? _controller;

  @override
  void initState() {
    super.initState();
    if (widget.textController == null) {
      _controller = TextEditingController(text: widget.initialValue);
    }
    _effectiveController!.addListener(_handleControllerChanged);
    if (widget.textType == TextType.password) {
      obscureText = true;
    }
  }

  @override
  void didUpdateWidget(TextFormInput oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.textController != oldWidget.textController) {
      oldWidget.textController?.removeListener(_handleControllerChanged);
      widget.textController?.addListener(_handleControllerChanged);

      if (oldWidget.textController != null && widget.textController == null)
        _controller =
            TextEditingController.fromValue(oldWidget.textController!.value);
      if (widget.textController != null) {
        setValue(widget.textController!.text);
        if (oldWidget.textController == null) _controller = null;
      }
    }
  }

  @override
  TextFormInput get widget => super.widget as TextFormInput;

  @override
  void dispose() {
    _effectiveController?.removeListener(_handleControllerChanged);
    _controller?.dispose();
    super.dispose();
  }

  void changeValue(String? v) {
    if (_effectiveController!.text != v) {
      _effectiveController!.text = v ?? '';
    }
    if (v != value) {
      didChange(v);
      if (widget.onChanged != null) {
        widget.onChanged!(v);
      }
    }
  }

  void setObscureText(bool obscureText) {
    setState(() {
      this.obscureText = obscureText;
    });
  }

  void _handleControllerChanged() {
    if (_effectiveController!.text != value)
      changeValue(_effectiveController!.text);
  }

  @override
  void reset() {
    super.reset();
    setState(() {
      _effectiveController!.text = widget.initialValue ?? '';
    });
  }

  @override
  bool validate() {
    saveValue();
    return super.validate();
  }

  @override
  void save() {
    saveValue();
    super.save();
  }

  void saveValue() {
    changeValue(_effectiveController!.text);
  }
}
